"""Init file for condoor."""

from condoor.connection import Connection
from condoor.config import CONF
from condoor.patterns import YPatternManager as PatternManager

from condoor.exceptions import CommandTimeoutError, ConnectionError, ConnectionTimeoutError, CommandError, \
    CommandSyntaxError, ConnectionAuthenticationError, GeneralError, InvalidHopInfoError, ConfigurationErrors, \
    ConfigurationSemanticErrors
from version import __version__

from pexpect import TIMEOUT, EOF


pattern_manager = PatternManager()

"""
This is a python module providing access to Cisco devices over Telnet and SSH.

"""

__all__ = ('Connection', 'TIMEOUT', 'EOF', 'pattern_manager', 'CONF', 'InvalidHopInfoError',
           'CommandTimeoutError', 'ConnectionError', 'ConnectionTimeoutError', 'CommandError',
           'CommandSyntaxError', 'ConnectionAuthenticationError', 'GeneralError', 'ConfigurationErrors',
           'ConfigurationSemanticErrors', '__version__')



"""This is IOS river implementation."""

from functools import partial
import re
import pexpect

from condoor.drivers.generic import Driver as Generic
from condoor.actions import a_send_password, a_expected_prompt, a_send_line, a_send, a_disconnect, a_reconnect
from condoor.exceptions import ConnectionAuthenticationError, ConnectionError, CommandError
from condoor.fsm import FSM


SAVE_CONFIG = re.compile(re.escape("System configuration has been modified. Save? [yes/no]: "))
PROCEED = re.compile(re.escape("Proceed with reload? [confirm]"))


class Driver(Generic):
    """This is a Driver class implementation for IOS."""

    platform = 'IOS'
    inventory_cmd = 'show inventory'
    users_cmd = 'show users'
    enable_cmd = 'enable'
    reload_cmd = 'reload'
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'rommon']
    prepare_terminal_session = ['terminal len 0', 'terminal width 0']
    families = {
        'A9': 'ASR900',
    }

    def __init__(self, device):
        """Initialize the IOS Driver object."""
        super(Driver, self).__init__(device)

    def get_version_text(self):
        """Return the version information from IOS device."""
        version_text = None
        try:
            version_text = self.device.send("show version", timeout=120)
        except CommandError as exc:
            exc.command = 'show version'
            raise exc

        return version_text

    def enable(self, enable_password):
        """Change to the privilege mode."""
        if self.device.prompt[-1] == '#':
            self.log("Device is already in privileged mode")
            return

        events = [self.password_re, self.device.prompt_re, pexpect.TIMEOUT, pexpect.EOF]
        transitions = [
            (self.password_re, [0], 1, partial(a_send_password, enable_password), 10),
            (self.password_re, [1], -1, ConnectionAuthenticationError("Incorrect enable password",
                                                                      self.device.hostname), 0),
            (self.device.prompt_re, [0, 1, 2, 3], -1, a_expected_prompt, 0),
            (pexpect.TIMEOUT, [0, 1, 2], -1, ConnectionAuthenticationError("Unable to get privileged mode",
                                                                           self.device.hostname), 0),
            (pexpect.EOF, [0, 1, 2], -1, ConnectionError("Device disconnected"), 0)
        ]
        self.device.ctrl.send_command(self.enable_cmd)
        fsm = FSM("IOS-ENABLE", self.device, events, transitions, timeout=10, max_transitions=5)
        fsm.run()
        if self.device.prompt[-1] != '#':
            raise ConnectionAuthenticationError("Privileged mode not set", self.device.hostname)

    def reload(self, reload_timeout=300, save_config=True):
        """Reload the device.

        CSM_DUT#reload

        System configuration has been modified. Save? [yes/no]: yes
        Building configuration...
        [OK]
        Proceed with reload? [confirm]
        """
        response = "yes" if save_config else "no"

        events = [SAVE_CONFIG, PROCEED, pexpect.TIMEOUT, pexpect.EOF]

        transitions = [
            (SAVE_CONFIG, [0], 1, partial(a_send_line, response), 60),
            (PROCEED, [0, 1], 2, partial(a_send, "\r"), reload_timeout),
            # if timeout try to send the reload command again
            (pexpect.TIMEOUT, [0], 0, partial(a_send_line, self.reload_cmd), 10),
            (pexpect.TIMEOUT, [2], -1, a_reconnect, 0),
            (pexpect.EOF, [0, 1, 2], -1, a_disconnect, 0)
        ]
        fsm = FSM("IOS-RELOAD", self.device, events, transitions, timeout=10, max_transitions=5)
        return fsm.run()



"""Provides the Controller class which is a wrapper to the pyexpect.spawn class."""

import re
import os
import pexpect
from time import time

from condoor.utils import delegate, levenshtein_distance
from condoor.exceptions import ConnectionError, ConnectionTimeoutError


# Delegate following methods to _session class
@delegate("_session", ("expect", "expect_exact", "expect_list", "compile_pattern_list", "sendline",
                       "isalive", "sendcontrol", "send", "read_nonblocking", "setecho", "delaybeforesend",
                       "waitnoecho"))
class Controller(object):
    """Controller class which wraps the pyexpect.spawn class."""

    def __init__(self, connection):
        """Initialize the Controller object for specific connection."""
        # delegated pexpect session
        self._session = None
        self._connection = connection

        self._logfile_fd = connection.session_fd
        self.connected = False
        self.authenticated = False
        self.last_hop = 0

    @property
    def hostname(self):
        """Return the hostname."""
        return self._connection.hostname

    def spawn_session(self, command):
        """Spawn the session using proper command."""
        if self._session and self.isalive():  # pylint: disable=no-member
            self._connection.log("Executing command: '{}'".format(command))
            try:
                self.send(command)  # pylint: disable=no-member
                self.expect_exact(command, timeout=20)  # pylint: disable=no-member
                self.sendline()  # pylint: disable=no-member

            except (pexpect.EOF, OSError):
                raise ConnectionError("Connection error", self.hostname)
            except pexpect.TIMEOUT:
                raise ConnectionTimeoutError("Timeout", self.hostname)

        else:
            self._connection.log("Spawning command: '{}'".format(command))
            env = os.environ
            env['TERM'] = 'vt100'  # to avoid color control characters and make sure other env not intact, esp. PATH
            try:
                self._session = pexpect.spawn(
                    command,
                    maxread=65536,
                    searchwindowsize=4000,
                    env=env,
                    echo=True  # KEEP YOUR DIRTY HANDS OFF FROM ECHO!
                )
                self._session.delaybeforesend = 0.3
                self._connection.log("Child process FD: {}".format(self._session.child_fd))
                rows, cols = self._session.getwinsize()
                if cols < 180:
                    self._session.setwinsize(512, 240)
                    nrows, ncols = self._session.getwinsize()
                    self._connection.log("Terminal window size changed from {}x{} to {}x{}".format(
                        rows, cols, nrows, ncols))
                else:
                    self._connection.log("Terminal window size: {}x{}".format(rows, cols))

            except pexpect.EOF:
                raise ConnectionError("Connection error", self.hostname)
            except pexpect.TIMEOUT:
                raise ConnectionTimeoutError("Timeout", self.hostname)

            self.set_session_log(self._logfile_fd)
            self.connected = True

    def set_session_log(self, session_log_fd=None):
        """Set the fd for session log."""
        self._connection.log("Setting the session log")
        if self._session:
            self._session.logfile_read = session_log_fd

    def send_command(self, cmd, password=False):
        """Send command."""
        try:
            if password:
                timeout = 10
                self._connection.log("Waiting for ECHO OFF")
                if self.waitnoecho(timeout):  # pylint: disable=no-member
                    self._connection.log("Password ECHO OFF received")
                else:
                    self._connection.log("Password ECHO OFF not received within {}s".format(timeout))
                self.sendline(cmd)  # pylint: disable=no-member
            else:
                self.send(cmd)  # pylint: disable=no-member
                self.expect_exact([cmd, pexpect.TIMEOUT], timeout=15)  # pylint: disable=no-member
                self.sendline()  # pylint: disable=no-member
        except OSError:
            self._connection.log("Session already disconnected.")
            raise ConnectionError("Session already disconnected")

    def disconnect(self):
        """Disconnect the controller."""
        if self._session and self._session.isalive():
            self._connection.log("Disconnecting the sessions")
            # write the session log buffer to disk
            self._session.logfile_read.write("\r\n")
            self._session.close(force=True)
            self._session.wait()
        self._connection.log("Disconnected")
        self.connected = False

    def try_read_prompt(self, timeout_multiplier):
        """Read the prompt.

        Based on try_read_prompt from pxssh.py
        https://github.com/pexpect/pexpect/blob/master/pexpect/pxssh.py
        """
        # maximum time allowed to read the first response
        first_char_timeout = timeout_multiplier * 2

        # maximum time allowed between subsequent characters
        inter_char_timeout = timeout_multiplier * 0.4

        # maximum time for reading the entire prompt
        total_timeout = timeout_multiplier * 4

        prompt = ""
        begin = time()
        expired = 0.0
        timeout = first_char_timeout

        while expired < total_timeout:
            try:
                char = self.read_nonblocking(size=1, timeout=timeout)  # pylint: disable=no-member
                # \r=0x0d CR \n=0x0a LF
                if char not in ['\n', '\r']:  # omit the cr/lf sent to get the prompt
                    timeout = inter_char_timeout
                expired = time() - begin
                prompt += char
            except pexpect.TIMEOUT:
                break
            except pexpect.EOF:
                raise ConnectionError('Session disconnected')

        prompt = prompt.strip()
        return prompt

    def detect_prompt(self, sync_multiplier=4):
        """Detect the prompt.

        This attempts to find the prompt. Basically, press enter and record
        the response; press enter again and record the response; if the two
        responses are similar then assume we are at the original prompt.
        This can be a slow function. Worst case with the default sync_multiplier
        can take 16 seconds. Low latency connections are more likely to fail
        with a low sync_multiplier. Best case sync time gets worse with a
        high sync multiplier (500 ms with default).

        """
        self.sendline()  # pylint: disable=no-member
        self.try_read_prompt(sync_multiplier)

        attempt = 0
        max_attempts = 10
        while attempt < max_attempts:
            attempt += 1
            self._connection.log("Detecting prompt. Attempt ({}/{})".format(attempt, max_attempts))

            self.sendline()  # pylint: disable=no-member
            first = self.try_read_prompt(sync_multiplier)

            self.sendline()  # pylint: disable=no-member
            second = self.try_read_prompt(sync_multiplier)

            lhd = levenshtein_distance(first, second)
            len_first = len(first)
            self._connection.log("LD={},MP={}".format(lhd, sync_multiplier))
            sync_multiplier *= 1.2
            if len_first == 0:
                continue

            if float(lhd) / len_first < 0.3:
                prompt = second.splitlines(True)[-1]
                self._connection.log("Detected prompt: '{}'".format(prompt))
                compiled_prompt = re.compile("(\r\n|\n\r){}".format(re.escape(prompt)))
                self.sendline()  # pylint: disable=no-member
                self.expect(compiled_prompt)  # pylint: disable=no-member
                return prompt

        return None

    @property
    def is_connected(self):
        """Return the session state regardless of device connection state."""
        return self.connected and self._session and self._session.isalive()

    @property
    def before(self):
        """Return text up to the expected string pattern."""
        return self._session.before if self._session else None

    @property
    def after(self):
        """Return text that was matched by the expected pattern."""
        return self._session.after if self._session else None



"""Provides the Chain class keeping the information about intermediate devices (jumphosts) on the paths to target."""
import re
import logging

from condoor.device import Device
from condoor.hopinfo import make_hop_info_from_url
from condoor.controller import Controller
from condoor.protocols import make_protocol
from condoor.exceptions import ConnectionError, CommandSyntaxError, CommandError


def device_gen(chain, urls):
    """Device object generator."""
    itr = iter(urls)
    last = next(itr)
    for url in itr:
        yield Device(chain, make_hop_info_from_url(last), driver_name='jumphost', is_target=False)
        last = url
    yield Device(chain, make_hop_info_from_url(last), driver_name='generic', is_target=True)


class Chain(object):
    """Chain class keeping information about the intermediate jumphosts and target device."""

    def __init__(self, connection, urls):
        """Initialize the new Chain object."""
        self.connection = connection
        self.ctrl = Controller(connection)
        self.devices = [device for device in device_gen(self, urls)]
        self.target_device.driver_name = 'generic'
        self.target_device.is_target = True

    def __repr__(self):
        """Return the string representation of devices in the chain."""
        name = ""
        for device in self.devices:
            name += "->{}".format(str(device))
        return name[2:]

    def connect(self):
        """Connect to the target device using the intermediate jumphosts."""
        device = None
        # logger.debug("Connecting to: {}".format(str(self)))
        for device in self.devices:
            if not device.connected:
                self.connection.emit_message("Connecting {}".format(str(device)), log_level=logging.INFO)
                protocol_name = device.get_protocol_name()
                device.protocol = make_protocol(protocol_name, device)
                command = device.protocol.get_command()
                self.ctrl.spawn_session(command=command)
                try:
                    result = device.connect(self.ctrl)
                except CommandSyntaxError as exc:
                    # all commands during discovery provides itself in exception except
                    # spawn session which is handled differently. If command syntax error is detected during spawning
                    # a new session then the problem is either on jumphost or csm server.
                    # The problem with spawning session is detected in connect FSM for telnet and ssh.
                    if exc.command:
                        cmd = exc.command
                        host = device.hostname
                    else:
                        cmd = command
                        host = "Jumphost/CSM"

                    self.connection.log(
                        "Command not supported or not authorized on {}: '{}'".format(host, cmd))
                    raise CommandError(message="Command not supported or not authorized",
                                       command=cmd,
                                       host=host)

                if result:
                    # logger.info("Connected to {}".format(device))
                    self.connection.emit_message("Connected {}".format(device), log_level=logging.INFO)
                else:
                    if device.last_error_msg:
                        message = device.last_error_msg
                        device.last_error_msg = None
                    else:
                        message = "Connection error"

                    self.connection.log(message)
                    raise ConnectionError(message)  # , host=str(device))

        if device is None:
            raise ConnectionError("No devices")

        return True

    def disconnect(self):
        """Disconnect from the device."""
        self.target_device.disconnect()
        self.ctrl.disconnect()
        self.tail_disconnect(-1)

    @property
    def target_device(self):
        """Return the target device object (last) in the chain."""
        try:
            return self.devices[-1]
        except IndexError:
            return None

    @property
    def is_connected(self):
        """Return if target device is connected."""
        if self.ctrl and self.ctrl.is_connected:
            # FIXME: Walk through device and check is_connected
            return True
        else:
            return False

    @property
    def is_discovered(self):
        """Return if target device is discovered."""
        if self.target_device is None:
            return False

        if None in (self.target_device.version_text, self.target_device.os_type, self.target_device.os_version,
                    self.target_device.inventory_text, self.target_device.family, self.target_device.platform):
            return False
        return True

    @property
    def is_console(self):
        """Return is target device is connected over console."""
        if self.target_device is None:
            return False

        return self.target_device.is_console

    def get_previous_prompts(self, device):
        """Return the list of intermediate prompts. All except target."""
        device_index = self.devices.index(device)
        prompts = [re.compile("(?!x)x")] + \
                  [dev.prompt_re for dev in self.devices[:device_index] if dev.prompt_re is not None]
        return prompts

    def get_device_index_based_on_prompt(self, prompt):
        """Return the device index in the chain based on prompt."""
        conn_info = ""
        for device in self.devices:
            conn_info += str(device) + "->"
            if device.prompt == prompt:
                self.connection.log("Connected: {}".format(conn_info))
                return self.devices.index(device)
        else:
            return None

    def tail_disconnect(self, index):
        """Mark all devices disconnected except target in the chain."""
        try:
            for device in self.devices[index + 1:]:
                device.connected = False
        except IndexError:
            pass

    def send(self, cmd, timeout, wait_for_string, password):
        """Send command to the target device."""
        return self.target_device.send(cmd, timeout=timeout, wait_for_string=wait_for_string, password=password)

    def update(self, data):
        """Update the chain object with the predefined data."""
        if data is None:
            for device in self.devices:
                device.clear_info()
        else:
            for device, device_info in zip(self.devices, data):
                device.device_info = device_info
                self.connection.log("Device information updated -> [{}]".format(device))



#!/usr/bin/env python
"""The condoor command line implementation."""

try:
    import click
except ImportError:
    print("Install click python package\n pip install click")
    exit()

import logging
import urlparse

import condoor


def echo_info(conn):
    """Print detected information."""
    click.echo("General information:")
    click.echo(" Hostname: {}".format(conn.hostname))
    click.echo(" HW Family: {}".format(conn.family))
    click.echo(" HW Platform: {}".format(conn.platform))
    click.echo(" SW Type: {}".format(conn.os_type))
    click.echo(" SW Version: {}".format(conn.os_version))
    click.echo(" Prompt: {}".format(conn.prompt))
    click.echo(" Console connection: {}".format(conn.is_console))

    click.echo("\nUDI:")
    click.echo(" PID: {}".format(conn.pid))
    click.echo(" Description: {}".format(conn.description))
    click.echo(" Name: {}".format(conn.name))
    click.echo(" SN: {}".format(conn.sn))
    click.echo(" VID: {}".format(conn.vid))


class URL(click.ParamType):
    """This is URL type validator."""

    name = 'url'

    def convert(self, value, param, ctx):
        """Convert to URL scheme."""
        if not isinstance(value, tuple):
            parsed = urlparse.urlparse(value)
            if parsed.scheme not in ('telnet', 'ssh'):
                self.fail('invalid URL scheme (%s).  Only telnet and ssh URLs are '
                          'allowed' % parsed, param, ctx)
        return value


log_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "ERROR": logging.ERROR,
    "NONE": logging.NOTSET,
}


@click.command()
@click.option("--url", multiple=True, required=True, envvar='CONDOOR_URLS', type=URL(),
              help='The connection url to the host (i.e. telnet://user:pass@hostname). '
                   'The --url option can be repeated to define multiple jumphost urls. '
                   'If no --url option provided the CONDOOR_URLS environment variable is used.')
@click.option("--log-path", default=None, type=click.Path(),
              help="The logging path. If no path specified condoor logs are sent to stdout and session logs "
                   "are sent to stderr.")
@click.option("--log-level", type=click.Choice(["NONE", "DEBUG", "INFO", "ERROR"]),
              show_default=True, default='ERROR',
              help='Logging level.')
@click.option("--log-session", is_flag=True,
              help="Log terminal session.")
@click.option("--force-discovery", is_flag=True,
              help="Force full device discovery.")
@click.option("--cmd", multiple=True, default=[],
              help='The command to be send to the device. The --cmd option can be repeated multiple times.')
@click.option("--print-info", is_flag=True,
              help="Print the discovered information.")
def run(url, cmd, log_path, log_level, log_session, force_discovery, print_info):
    """Run the main function."""
    log_level = log_levels[log_level]
    conn = condoor.Connection("host", list(url), log_session=log_session, log_level=log_level, log_dir=log_path)
    try:
        conn.connect(force_discovery=force_discovery)
        if print_info:
            echo_info(conn)

        for command in cmd:
            result = conn.send(command)
            print("\nCommand: {}".format(command))
            print("Result: \n{}".format(result))
    except (condoor.ConnectionError, condoor.ConnectionAuthenticationError, condoor.ConnectionTimeoutError,
            condoor.InvalidHopInfoError, condoor.CommandSyntaxError, condoor.CommandTimeoutError,
            condoor.CommandError, condoor.ConnectionError) as excpt:
        click.echo(excpt)
    finally:
        conn.disconnect()
    return


if __name__ == '__main__':
    run()  # pylint: disable=no-value-for-parameter



#! /usr/bin/python
# =============================================================================
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

"""
 Installation script for accelerated upgrade
"""
import codecs
import re
import sys

from setuptools.command.test import test as TestCommand

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup



DESCRIPTION = 'This is a python module providing access to Cisco devices over Telnet and SSH'
with codecs.open('README.md', 'r', encoding='UTF-8') as readme:
    LONG_DESCRIPTION = ''.join(readme)

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: POSIX :: Linux',
]

packages = [
    'condoor',
    'condoor.protocols',
    'condoor.drivers',
]

NAME = 'condoor'


def version():
    pyfile = 'condoor/version.py'
    with open(pyfile) as fp:
        data = fp.read()

    match = re.search("__version__ = '([^']+)'", data)
    assert match, 'cannot find version in {}'.format(pyfile)
    return match.group(1)


class Tox(TestCommand):
    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #  import here, cause outside the eggs aren't loaded
        import tox
        import shlex
        args = self.tox_args
        if args:
            args = shlex.split(self.tox_args)
        errno = tox.cmdline(args=args)
        sys.exit(errno)


setup(
    name=NAME,
    version=version(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author='Klaudiusz Staniek',
    author_email='klstanie@cisco.com',
    url='https://github.com/kstaniek/condoor',
    download_url='https://github.com/kstaniek/condoor/tarball/{}'.format(version()),
    keywords='cisco,automation',
    tests_require=['tox'],
    cmdclass={'test': Tox},
    platforms=['any'],
    packages=packages,
    package_data={'': ['LICENSE', ], },
    package_dir={'condoor': 'condoor'},
    include_package_data=True,
    install_requires=['pexpect>=4.2.1', 'pyyaml'],
    data_files=[('condoor', ['condoor/patterns.yaml', 'condoor/config.yaml'])],
    license='Apache 2.0',
    classifiers=CLASSIFIERS,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'condoor = condoor.__main__:run',
        ]
    }
)
